@@ -12,7 +12,7 @@
 #endif
 #include "system.h"
 
-int timescmd() {
+int timescmd(int argc, char *argv[]) {
 	struct tms buf;
 	long int clk_tck = sysconf(_SC_CLK_TCK);
 	int mutime, mstime, mcutime, mcstime;
